library Stat;

uses Objects, WinAPI;

{$IFNDEF Windows}
  This DLL must be compiled for Windows!
{$ENDIF}

type
  TStatResult = (srOK, srInvalidContext, srNoData, srNoMemory, srMaxSamples,
    srInvalidSample, srError, srNoStandard);
  TCalcResult = (crMean, crSum, crSumOfSquares, crStandard, crPopulation);

  PSample = ^TSample;
  TSample = record
    Sample: Real;
  end;

  TGetMem = procedure(var P; Size: Word);
  TFreeMem = procedure(var P; Size: Word);

  PStatCollection = ^TStatCollection;
  TStatCollection = object(TCollection)
    ErrorCode: Integer;
    GetMem: TGetMem;
    FreeMem: TFreeMem;
    constructor Init(ASize, ADelta: Integer; AGetMem: TGetMem;
      AFreeMem: TFreeMem);
    procedure Error(Code, Info: Integer); virtual;
    procedure FreeItem(Item: Pointer); virtual;
    function Mean(var R): TStatResult;
    procedure SetLimit(ALimit: Integer); virtual;
    function Sum(var R): TStatResult;
    function SumOfSquares(var R): TStatResult;
    function StandardDeviation(var R): TStatResult;
    function PopulationDeviation(var R): TStatResult;
  end;

  TStatContext = PStatCollection;

function InitCalc(Size, Delta: Integer; GetMem: TGetmem; FreeMem: TFreeMem):
  TStatContext; export;
var
  Calc: PStatCollection;
begin
  GetMem(Calc, SizeOf(TStatCollection));
  Calc^.Init(Size, Delta, GetMem, FreeMem);
  InitCalc := Calc;
end;

function AddSample(Context: TStatContext; const Sample: TSample):
  TStatResult; export;
var
  ASample: PSample;
begin
  if Context = nil then
    AddSample := srInvalidContext
  else with PStatCollection(Context)^ do
  begin
    GetMem(ASample, SizeOf(TSample));
    ASample^ := Sample;
    Insert(ASample);
    if ErrorCode <> 0 then
      AddSample := srNoMemory
    else AddSample := srOK;
    ErrorCode := 0;
  end;
end;

function DeleteSample(Context: TStatContext; Item: Integer):
  TStatResult; export;
begin
  if Context = nil then
    DeleteSample := srInvalidContext
  else with PStatCollection(Context)^ do
  begin
    AtFree(Item);
    if ErrorCode <> 0 then
      DeleteSample := srInvalidSample
    else DeleteSample := srOK;
    ErrorCode := 0;
  end;
end;

procedure DoneCalc(Context: TStatContext); export;
begin
  if Context = nil then Exit;
  with PStatCollection(Context)^ do
  begin
    Done;
    FreeMem(Context, SizeOf(TStatCollection));
  end;
end;

function GetSample(Context: TStatContext; Item: Integer;
  var Sample: PSample): TStatResult; export;
begin
  if Context = nil then
    GetSample := srInvalidContext
  else with PStatCollection(Context)^ do
  begin
    Sample := At(Item);
    if ErrorCode <> 0 then
      GetSample := srInvalidSample
    else GetSample := srOK;
    ErrorCode := 0;
  end;
end;

function SampleCount(Context: TStatContext; var ACount: Integer): TStatResult; export;
begin
  SampleCount := srOK;
  if Context = nil then
    SampleCount := srInvalidContext
  else with PStatCollection(Context)^ do
    ACount := Count;
end;

function GetCalc(Context: TStatContext; Calc: TCalcResult;
  var Rslt): TStatResult; export;
begin
  GetCalc := srOK;
  if Context = nil then
    GetCalc := srInvalidContext
  else with PStatCollection(Context)^ do
    case Calc of
      crMean: GetCalc := Mean(Rslt);
      crSum: GetCalc := Sum(Rslt);
      crSumOfSquares: GetCalc := SumOfSquares(Rslt);
      crStandard: GetCalc := StandardDeviation(Rslt);
      crPopulation: GetCalc := PopulationDeviation(Rslt);
    else
      GetCalc := srError;
    end;
end;

{ Be sure to compile this for Windows since this constructor assumes the
  fields are not cleared by TObject. }

constructor TStatCollection.Init(ASize, ADelta: Integer; AGetMem: TGetMem;
  AFreeMem: TFreeMem);
begin
  GetMem := AGetMem;
  FreeMem := AFreeMem;
  inherited Init(ASize, ADelta);
  ErrorCode := 0;
end;

procedure TStatCollection.Error(Code, Info: Integer);
begin
  ErrorCode := Code;
end;

procedure TStatCollection.FreeItem(Item: Pointer);
begin
  FreeMem(Item, SizeOf(TSample));
end;

function TStatCollection.Mean(var R): TStatResult;
var
  ASum: Real;
begin
  Mean := srNoData;
  if Count = 0 then Exit;
  Sum(ASum);
  Real(R) := ASum / Count;
  Mean := srOK;
end;

{ The built-in TCollection automatically calls GetMem and FreeMem when
  changing the size of the collection.  The problem is that the DLL's
  heap routines are called instead of the client's.  By duplicating
  SetLimit, the calls to the heap managaer can be moved into the correct
  scope so the client's heap functions are called. }

procedure TStatCollection.SetLimit(ALimit: Integer);
var
  AItems: PItemList;
begin
  if ALimit < Count then ALimit := Count;
  if ALimit > MaxCollectionSize then ALimit := MaxCollectionSize;
  if ALimit <> Limit then
  begin
    if ALimit = 0 then AItems := nil else
    begin
      GetMem(AItems, Alimit * SizeOf(Pointer));
      if (Count <> 0) and (Items <> nil) then
        Move(Items^, AItems^, Count * SizeOf(Pointer));
    end;
    if Limit <> 0 then FreeMem(Items, Limit * SizeOf(Pointer));
    Limit := ALimit;
    Items := AItems;
  end;
end;

function TStatCollection.Sum(var R): TStatResult;
var
  ASum: Real;

  procedure DoSum(var S: TSample); far;
  begin
    ASum := ASum + S.Sample;
  end;

begin
  Sum := srNoData;
  if Count = 0 then Exit;
  ASum := 0.0;
  ForEach(@DoSum);
  Real(R) := ASum;
  Sum := srOK;
end;

function TStatCollection.SumOfSquares(var R): TStatResult;
var
  ASumOfSquares: Real;

  procedure DoSumOfSquares(var S: TSample); far;
  begin
    ASumOfSquares := ASumOfSquares + S.Sample * S.Sample;
  end;

begin
  SumOfSquares := srNoData;
  if Count = 0 then Exit;
  ASumOfSquares := 0.0;
  ForEach(@DoSumOfSquares);
  Real(R) := ASumOfSquares;
  SumOfSquares := srOK;
end;

function TStatCollection.StandardDeviation(var R): TStatResult;
var
  AMean, ASumOfSquares: Real;
begin
  StandardDeviation := srNoStandard;
  if Count < 2 then Exit;
  Mean(AMean);
  SumOfSquares(ASumOfSquares);
  Real(R) := Sqrt((ASumOfSquares - (Count * (AMean * AMean))) / (Count - 1));
  StandardDeviation := srOK;
end;

function TStatCollection.PopulationDeviation(var R): TStatResult;
var
  AMean, ASumOfSquares: Real;
begin
  PopulationDeviation := srNoData;
  if Count = 0 then Exit;
  Mean(AMean);
  SumOfSquares(ASumOfSquares);
  Real(R) := Sqrt((ASumOfSquares - (Count * (AMean * AMean))) / Count);
  PopulationDeviation := srOK;
end;

exports
  InitCalc,
  DoneCalc,
  AddSample,
  DeleteSample,
  GetSample,
  SampleCount,
  GetCalc;

begin
end.